// #![no_std]
#[macro_use]
extern crate serde_derive;
extern crate serde;
extern crate serde_json;
use mentat::MonotonicCubicSpline;
// use splines::{Interpolation, Key, Spline};
use std::collections::HashMap;
use std::fs::File;
use std::io::BufReader;
use std::io::BufWriter;

#[derive(Serialize, Deserialize, Debug)]
pub enum Adjustment {
    MOAQuarter = 4,
    MOAHalf = 2,
}

impl Default for Adjustment {
    fn default() -> Self {
        Adjustment::MOAQuarter
    }
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Gun {
    pub name: String,
    pub adjustment: Adjustment,
    pub caliber: f32,
    #[serde(default)]
    pub weight: f32,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct BallisticsData {
    pub velocity: Vec<(u32, f32)>,
    pub energy: Vec<(u32, f32)>,
    pub trajectory: Vec<(u32, f32)>,
    #[serde(default)]
    pub diameter: f64,
    #[serde(default)]
    pub trajectory_filled: Vec<(u32, f32)>,
}

impl BallisticsData {

    pub fn monospline(&self) -> MonotonicCubicSpline {
        let x: Vec<f64> = self.trajectory.iter().map(|x| x.0 as f64).collect();
        let y: Vec<f64> = self.trajectory.iter().map(|x| x.1 as f64).collect();
        MonotonicCubicSpline::new(&x, &y)
    }

    fn data_at(&self, d: u32) {
        // let vel = self.attr_at(d, &self.velocity);
        // let energy = self.attr_at(d, &self.energy);
        // let trajectory = self.attr_at(d, &self.trajectory);
        // println!("D{} V{} E{} T{}", d, vel, energy, trajectory);
    }


    fn interpolate_drop_at(&self, dist: u32) -> f64 {
        self.monospline()
            .interpolate(dist as f64)
    }

    pub fn drop_zeroed(&self, dist: u32, zero_dist: u32) -> f64 {
        let z_comp = self
            .interpolate_drop_at(zero_dist);
        // println!("drop at {} {}", dist, z_comp);
        self.interpolate_drop_at(dist) - z_comp
    }
}

fn moa_at_distance(d: u32) -> f64 {
    let radius: f64 = d as f64;
    let diameter = 2. * radius * std::f64::consts::PI;
    let deg = diameter/360.;
    let moa = deg/60.*100.;
    dbg!(&moa);
    // 1
    moa
}

fn clicks_at_distance(distance: u32, drop: f64, moa_divisor: Adjustment) -> i32 {
    (drop / ( moa_at_distance(distance)  / (moa_divisor as i32) as f64 )) as i32
}


fn _normalized_pos_in_range(min: f32, max: f32, d: f32) -> f32 {
    (d - min) / (max - min)
}
fn _lerp(y1: f32, y2: f32, mu: f32) -> f32 {
    y1 * (1.0 - mu) + y2 * mu
}
fn _write_sample() {
    let mut bd = BallisticsData::default();
    bd.velocity.push((0, 720.0));
    let mut map = HashMap::new();
    map.insert("Hammerhead", bd);
    // write out the file
    let writer = BufWriter::new(File::create("sample.json").unwrap());
    serde_json::to_writer_pretty(writer, &map).unwrap();
}

pub fn load_cartridges() -> Result<HashMap<String, BallisticsData>, serde_json::Error> {
    let reader = BufReader::new(File::open("loading_data.json").unwrap());
    serde_json::from_reader(reader)
}

#[test]
fn test_drop() {
    //write_sample();

    match load_cartridges() {
        Ok(c) => {
            for (name, cartridge) in c {
                println!("=== Cartridge {} ===", name);
                let distances = vec![100, 200];

                for d in distances {
                    dbg!(&d);

                    let drop = cartridge.interpolate_drop_at(d);
                    
                    dbg!(cartridge.drop_zeroed(d, 100));
                    dbg!(&drop);
                    let clicks = clicks_at_distance(d, drop, Adjustment::MOAQuarter);
                    dbg!(&clicks);

                }

       
            }
        }
        Err(e) => println!("{:?}", e),
    }
}


#[cfg(test)]
use plotlib::page::Page;
#[cfg(test)]
use plotlib::repr::Scatter;
#[cfg(test)]
use plotlib::style::{PointMarker, PointStyle};
#[cfg(test)]
use plotlib::view::ContinuousView;
#[test]
fn plot_graphs(){

    match load_cartridges() {
        Ok(c) => {
            for (name, cartridge) in c {
                let mut data = vec![];
                for d in 50..300 {
                    let r = cartridge.interpolate_drop_at(d);
                    data.push((d as f64, r as f64));
                }

                // We create our scatter plot from the data
                let s1: Scatter = Scatter::from_slice(&data).style(
                    PointStyle::new()
                        .size(1.1)
                );
                // The 'view' describes what set of data is drawn
                let v = ContinuousView::new()
                    .add(s1)
                    .x_range(0., 300.)
                    .y_range(-100., 20.)
                    .x_label("Distance")
                    .y_label(format!("Drop for {}", name));

                // A page with a single view is then saved to an SVG file
                Page::single(&v).save(format!("plot-{}.svg", name)).unwrap();
            }
        }
        Err(e) => println!("{:?}", e),
    }

}